/*
 * videodirlinked.c: Functions to maintain the video directory
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <vdr/recording.h>
#include <vdr/tools.h>
#include "videodirlinked.h"
#include "videodirtools.h"

#if (APIVERSNUM >= 20102)

#define SORTMODEFILE      ".sort"

// #undef  dsyslog
// #define dsyslog esyslog


// --- cVideoDirectoryLinked -------------------------------------------------------


bool cVideoDirectoryLinked::MoveNRelink(const char *OldName, const char *NewName) {
   dsyslog("MoveNRelink Recording: %s -> %s", OldName, NewName);
   //int Name_length = strlen(Name());
   int OldName_length = strlen(OldName);
   int NewName_length = strlen(NewName);
   
   
   cReadDir d(OldName);
   struct dirent *e;
   while ((e = d.Next()) != NULL) {
      char *LinkName = e->d_name;
      int LinkName_length = strlen(LinkName);
      char OldSourceName[OldName_length + LinkName_length + 5];
      sprintf(OldSourceName, "%s/%s", OldName, LinkName);
      struct stat st;
      if (lstat(OldSourceName, &st) == 0) {
         if (S_ISLNK(st.st_mode)) {
            dsyslog("MoveNRelink start %s", OldSourceName);
            int OldTargetName_length = st.st_size + 5; // buf length!
            char OldTargetName[OldTargetName_length];
            int tmp01 = readlink(OldSourceName, OldTargetName, OldTargetName_length);
            if((tmp01 == -1) || (tmp01 > OldTargetName_length - 2)) {
               LOG_ERROR_STR(OldTargetName);
               return false;
            }
            OldTargetName[tmp01] = 0;
            char NewSourceName[NewName_length + LinkName_length + 5];
            sprintf(NewSourceName, "%s/%s", NewName, LinkName);
            const char *NewTargetName = NULL;
            cLinkWalker Dirs;
            while (Dirs.Next()) {
               // try to find the filesystem of the LinkTarget.
               if (strstr(OldTargetName, Dirs.Name()) == OldTargetName) {
                  dsyslog("Equal beginning: %s  %s", Dirs.Name(),OldTargetName);
                  Dirs.Store();
                  NewTargetName = Dirs.Adjust(NewSourceName);
                  break;
               }
            }
            if (NewTargetName == NULL) {
               // alternative Attempt to find the filesystem of the LinkTarget.
               // this should work even if something has been wired linked, but it may spinup all disks.
               Dirs.Clear();
               while (Dirs.Next()) {
                  if (EntriesOnSameFileSystem(Dirs.Name(), OldTargetName)) {
                     dsyslog("EntriesOnSameFileSystem: %s  %s", Dirs.Name(),OldTargetName);
                     Dirs.Store();
                     NewTargetName = Dirs.Adjust(NewSourceName);
                     break;
                  }
               }
                if (NewTargetName == NULL) { 
                  esyslog("ERROR: Unable to find new name for: %s", OldSourceName);
                  return false;
                }
            }
            
            char OldTargetDir[strlen(OldTargetName) + 2];
            strcpy(OldTargetDir, OldTargetName);
            char *OldTargetDir_File = strrchr(OldTargetDir, '/');
            OldTargetDir_File[0] = 0;
            char NewTargetDir[strlen(NewTargetName) + 2];
            strcpy(NewTargetDir, NewTargetName);
            char *NewTargetDir_File = strrchr(NewTargetDir, '/');
            NewTargetDir_File[0] = 0;
            isyslog("Oldlink: %s -> %s", OldSourceName, OldTargetName);
            isyslog("Newlink: %s -> %s", NewSourceName, Dirs.LinksaveName());
            dsyslog("Dirs: %s  %s", OldTargetDir, NewTargetDir);  
            dsyslog("NewTargetname: %s", NewTargetName);     
            
            if (!MakeDirs(NewTargetDir, true)) {
               esyslog("ERROR: Could not create directory: %s", NewTargetDir);
               return false; // errno has been set by MakeDirs()
            }
//             if(!DirectoryOk(NewTargetDir, true)) {
//                LOG_ERROR_STR(NewTargetDir);
//                return false;
//             } 
            if(unlink(OldSourceName) < 0) {
               LOG_ERROR_STR(OldSourceName);
               return false;
            }
            if(strcmp(NewSourceName, NewTargetName) != 0) {
               if (symlink(Dirs.LinksaveName(), OldSourceName) < 0) { // OldSourceName! 
                     LOG_ERROR_STR(Dirs.LinksaveName());
                     return false;
               }
            }
            errno = 0;
            dsyslog("Rename Targetfile: %s > %s", OldTargetName, NewTargetName);
            if (rename(OldTargetName, NewTargetName) == -1) {
               LOG_ERROR_STR(NewTargetName);
               return false;
            }
            const char *IgnoreFiles[] = { SORTMODEFILE, NULL };
            RemoveEmptyDirectories(OldTargetDir, true, IgnoreFiles);
            dsyslog("MoveNRelink done: %s", NewTargetName);
         }
      } else {
         esyslog("ERROR: Could not stat file: %s", OldSourceName);
         return false;
      }
   }
   dsyslog("MoveNRelink finished Recording: %s", NewName);
   return true;
}


      ///< Returns the total amount (in MB) of free disk space for recording.
      ///< If UsedMB is given, it returns the amount of disk space in use by
      ///< existing recordings (or anything else) on that disk.
int cVideoDirectoryLinked::FreeMB(int *UsedMB)
{
  //dsyslog("VDLinked FreeMB");   //too noisy!
  
  int summ = 0;
  int u_summ = 0;
  cLinkWalker Dirs;
  while (Dirs.Next()) {
     summ += FreeDiskSpaceMB(Dirs.Name(), UsedMB);
     if (UsedMB)
        u_summ += *UsedMB;
  }
  
  if (UsedMB)
     *UsedMB = u_summ;

  //dsyslog("FreeMB at Videodir: %dMiB", summ);   //too noisy!
  return summ;
}

      ///< By default VDR assumes that the video directory consists of one large
      ///< volume, on which it can store its recordings. A derived cVideoDirectory
      ///< may, for instance, use several separate disks to store recordings.
      ///< The given FileName is the full path name (including the video directory) of
      ///< a recording file ('*.ts') that is about to be opened for writing. If the actual
      ///< file shall be put on an other disk, the derived cVideoDirectory should
      ///< create a symbolic link from the given FileName to the other location.
      ///< Returns true if the operation was successful.
      ///< The default implementation just checks whether the incoming file name really
      ///< is under the video directory.
bool cVideoDirectoryLinked::Register(const char *FileName)
{
   dsyslog("VDLinked Register %s", FileName);
   const char *LinkTargetName = NULL;
   
   // Incoming name must be in base video directory:
   if (strstr(FileName, Name()) != FileName) {
      esyslog("ERROR: %s not in %s", FileName, Name());
      errno = ENOENT; // must set 'errno' - any ideas for a better value?
      return false;
   }
   
   
   cLastExistingDir Lastdir(FileName);
   if (Lastdir.Name()) {
      if (!EntriesOnSameFileSystem(Name(), Lastdir.Name())) {
         //FileName is not on Video0. User has to take care about diskspace! 
         isyslog("Filename is not inside distributed Videodir (%s)! %s.", FileName, Name());
         return true;
      }
   } else {
      LOG_ERROR_STR(FileName);
      return false;
   }
   
   
   cLinkWalker Dirs;
   // Find the directory with the most free space:
   int MaxFree = 0;
   while (Dirs.Next()) {
      int Free = FreeDiskSpaceMB(Dirs.Name());
      dsyslog("FreeSpace at %s: %dMiB", Dirs.Name(), Free);
      if (Free > MaxFree) {
         Dirs.Store();
         MaxFree = Free;
      }
   }  
   
   LinkTargetName = Dirs.Adjust(FileName);
   if(strcmp(FileName, LinkTargetName) != 0) { // we have to create a symlink
      dsyslog("Register: Symlink to: %s", LinkTargetName);
      if (!MakeDirs(LinkTargetName, false)) {
         esyslog("ERROR: Could not create directory: %s", LinkTargetName);
         return false; // errno has been set by MakeDirs()
      }
      
      isyslog("Symlink: %s -> %s", FileName, Dirs.LinksaveName());
      if (symlink(Dirs.LinksaveName(), FileName) < 0) {
         if (errno == EEXIST) {
            isyslog("info: Symlink already existing: %s", FileName);
            // Default implementation will return true in this case too. 
         } else {
            esyslog("ERROR: Could not create symlink: %s -> %s", FileName, Dirs.LinksaveName());
            return false;
         }
      }
   } else
       isyslog("Register: %s = %s", FileName, LinkTargetName);
       
   
   dsyslog("Register: finished successfully");
   return true;
}

      ///< Renames the directory OldName to NewName.
      ///< OldName and NewName are full path names that begin with the name of the
      ///< video directory and end with '*.rec' or '*.del'. Only the base name (the
      ///< rightmost component) of the two names may be different.
      ///< Returns true if the operation was successful.
      ///< The default implementation just calls the system's rename() function.
bool cVideoDirectoryLinked::Rename(const char *OldName, const char *NewName)
{
  dsyslog("VDLinked: Rename %s to %s", OldName, NewName);
  
  if(EntriesOnSameFileSystem(OldName, Name())) {
     dsyslog("Rename: MoveNRelink");
     // Recording is on (possibly distributed) Videodir. Need to check for links. 
     if(!MoveNRelink(OldName, NewName))
        return false;
  }
  dsyslog("Rename: renaming basedir");
  isyslog("Renaming: %s to %s", OldName, NewName);
  if (rename(OldName, NewName) == -1) {
     LOG_ERROR_STR(NewName);
     return false;
  }
  
  dsyslog("Rename: finished successfully");
  return true;
}


      ///< Moves the directory FromName to the location ToName. FromName is the full
      ///< path name of a recording's '*.rec' directory. ToName has the same '*.rec'
      ///< part as FromName, but a different directory path above it.
      ///< Returns true if the operation was successful.
      ///< The default implementation just calls the system's rename() function.
bool cVideoDirectoryLinked::Move(const char *FromName, const char *ToName)
{
  dsyslog("VDLinked: moving %s to %s", FromName, ToName);
  
  if (EntriesOnSameFileSystem(FromName, Name())) {
     
     if(EntriesOnSameFileSystem(ToName, Name())) {
        // FromName and ToName are on distributed Videodir.
        isyslog("Moving inside Videodir.");
        if(!MoveNRelink(FromName, ToName)) {
           LOG_ERROR_STR(ToName);
           return false;
           }
        isyslog("Moving: %s to %s",FromName, ToName);
        if (rename(FromName, ToName) == -1) {
           LOG_ERROR_STR(ToName);
           return false;
        }
        return true;
        
     } // else: FromName is on distributed Videodir, ToName not. -> just move there
     
  } else {
     int symlinks_count = ContainsSymlinks(FromName);
     if(symlinks_count > 0) {
        esyslog("ERROR: No symlinked videofiles outside Videodir allowed! %s", FromName);
        return false;
     } else if (symlinks_count < 0)
        return false;
     
     if(EntriesOnSameFileSystem(ToName, Name())) {
     // FromName is not on distributed Videodir, but ToName is. -> need to find right directory for the files.
        isyslog("Moving to Videodir.");
        if ((DirSizeMB(FromName) + 10) >= (FreeMB(NULL))) {
           esyslog("ERROR: Insufficant diskspace on Videodir!");
           return false;
        }
         cReadDir d(FromName);
         struct dirent *e;
         while ((e = d.Next()) != NULL) {
            int i = 0;
            char *FileName = e->d_name;
            
            while(isdigit(FileName[i]))
               i++;
            
            if((i == 3) && (strstr(FileName, ".vdr") == (FileName + 3))) {
               isyslog("Old PES-recording (.vdr): %s/%s", ToName, FileName);
            } else if((i == 5) && (strstr(FileName, ".ts") == (FileName + 5))) {
               isyslog("New TS-recording (.ts): %s/%s", ToName, FileName);
            } else if(i != 0) {
               isyslog("No such filemame in recording expected!: %s/%s", ToName, FileName);
               continue;
            } else {
               continue;
            }
            char FullToFileName[strlen(ToName) + strlen(FileName) + 5];
            sprintf(FullToFileName, "%s/%s", ToName, FileName);
            if(!Register(FullToFileName)) {
               esyslog("ERROR: Could not register file: %s", FullToFileName);
               return false;
            }
         }
        return RecordingsHandler.Add(ruMove, FromName, ToName);
        
     } // else: Recording is not on distributed Videodir. -> just move
  }
  isyslog("Moving out of or outside Videodir.");
  
  if ((DirSizeMB(FromName) + 10) >= (FreeDiskSpaceMB(ToName))) {
     esyslog("ERROR: Insufficant diskspace on %s!", ToName);
     return false;
  }
  
  if (EntriesOnSameFileSystem(FromName, ToName)) {
     if (rename(FromName, ToName) == -1) {
        LOG_ERROR_STR(ToName);
        return false;
     }
  } else
     return RecordingsHandler.Add(ruMove, FromName, ToName);
  
  dsyslog("Move: finished successfully");
  return true;
}

      ///< Removes the directory with the given Name and everything it contains.
      ///< Name is a full path name that begins with the name of the video directory.
      ///< Returns true if the operation was successful.
      ///< The default implementation calls RemoveFileOrDir().
bool cVideoDirectoryLinked::Remove(const char *Name)
{  
  dsyslog("VDLinked Remove: %s", Name);
  return RemoveFileOrDir(Name, true); 
  // Changed "FollowSymlinks" to "true". TODO: Check if it is sufficient.
}

      ///< Recursively removes all empty directories under the video directory.
      ///< If IgnoreFiles is given, the file names in this (NULL terminated) array
      ///< are ignored when checking whether a directory is empty. These are
      ///< typically "dot files", like e.g. ".sort".
      ///< The default implementation calls RemoveEmptyDirectories().
void cVideoDirectoryLinked::Cleanup(const char *IgnoreFiles[])
{
  dsyslog("VDLinked Cleanup");
  
  cLinkWalker Dirs;
  while (Dirs.Next()) {
     RemoveEmptyDirectories(Dirs.Name(), false, IgnoreFiles);
  }
}

      ///< Checks whether the directory Name is on the same file system as the
      ///< video directory. Name is the full path name of a recording's '*.rec'
      ///< directory. This function is usually called when an ongoing recording
      ///< is about to run out of disk space, and an existing (old) recording needs
      ///< to be deleted. It shall make sure that deleting this old recording will
      ///< actually free up space in the video directory, and not on some other
      ///< device that just happens to be mounted.
      ///< The default implementation calls EntriesOnSameFileSystem().
bool cVideoDirectoryLinked::Contains(const char *Name)
{
  dsyslog("VDLinked Contains: %s", Name);
    
  cLinkWalker Dirs;
  while (Dirs.Next()) {
     if(EntriesOnSameFileSystem(Dirs.Name(), Name))
        return true;
  }
  
  return false;
}

void cVideoDirectoryLinked::LogInfo(void)
{
    int i=0;
    int summ = 0;
    int u_summ = 0;
    cLinkWalker Dirs;
    while (Dirs.Next()) {
        Dirs.Store();
        Dirs.Adjust(Dirs.Name());
        int freeMB=0;
        int usedMB=0;
        freeMB=FreeDiskSpaceMB(Dirs.Name(), &usedMB);
        summ += freeMB;
        u_summ += usedMB;
        isyslog("VDLinked #%d: free:%dMB used:%dMB %s > %s", i, freeMB, usedMB, Dirs.Name(), Dirs.LinksaveName());
        i++;
    }
    isyslog("VDLinked ->: free:%dMB used:%dMB, %d Disks, Basename: %s", summ, u_summ, i, cVideoDirectoryLinked::Name());
    if(cVideoDirectoryLinked::Name()[0] != '/')
        isyslog("Warning: Pathname for VideoDirectory isn't absolute. This may not work as expected!");
}

#endif 
