/*
 * videodirtools.c: tools for videodirlinked.c
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <vdr/tools.h>
#include <vdr/config.h>
#include "videodirlinked.h"
#include "videodirtools.h"

#if (APIVERSNUM >= 20102)

// #undef  dsyslog
// #define dsyslog esyslog



// --- cLinkWalker ----------------------------------------------------------

cLinkWalker::cLinkWalker(void)
{
   name = strdup(cVideoDirectoryLinked::Name());
   name_stored = NULL;
   filename_adjusted = NULL;
   linksave_filename = NULL;
   p_linksave_filename = (const char**) &filename_adjusted;
   length = strlen(name);
   number = 0;
   digits = 0;
   
   if (name[length-1] == '0') { 
      // The name of a distributed VideoDirectory must end with '0'!
      
      int l = length;
      while (l-- > 0 && isdigit(name[l]))
         ;
      l++;
      digits = length - l;
      //dsyslog("Potentially distributed VideoDirectory found.");   //too noisy!
      
      return;
   }
   // dsyslog("VideoDirectory not distributed.");   //too noisy!
};

cLinkWalker::~cLinkWalker()
{
  free(name);
  free(name_stored);
  free(filename_adjusted);
  free(linksave_filename);
};
   
bool cLinkWalker::Next(void)
{
  if (digits > 0) { // digits == 0 -> not distributed!
     char buf[16];
     if (sprintf(buf, "%0*d", digits, number) == digits) {
        strcpy(&name[length - digits], buf);
        number++;
        return DirectoryOk(name);
     }
  }
  return false;
};

const char *cLinkWalker::Name(void)
{
  return name;
};

const char *cLinkWalker::NameStored(void)
{
  return name_stored;
};

void cLinkWalker::Clear(void)
{
   free(name);
   name = strdup(cVideoDirectoryLinked::Name());
   free(name_stored);
   name_stored=NULL;
   free(filename_adjusted);
   filename_adjusted=NULL;
   free(linksave_filename);
   linksave_filename=NULL;
   p_linksave_filename = (const char**) &filename_adjusted;
   number = 0;
};

void cLinkWalker::Store(void)
{
     free(name_stored);
     name_stored = strdup(name);
};

const char *cLinkWalker::Adjust(const char *FileName)
{
  if (!name_stored)
     Store();
  
  free(filename_adjusted);
  filename_adjusted = strdup(FileName);
  if(name_stored[0] != '/') {
     // esyslog("Warning: Pathname for VideoDirectory isn't absolute. This isn't supported and may not work as expected!");   //too noisy!
     char *NameCanonical = NULL;
     NameCanonical = canonicalize_file_name(name_stored);
     if (NameCanonical) {
        linksave_filename = (char*)malloc(strlen(NameCanonical) + strlen(FileName) - length +5);
        strcpy(linksave_filename, NameCanonical);
        strcpy(&linksave_filename[strlen(NameCanonical)], &FileName[length]);
        p_linksave_filename = (const char**) &linksave_filename;
     } else {
        LOG_ERROR_STR(FileName);
        p_linksave_filename = &FileName;
     }
  }
  return strncpy(filename_adjusted, name_stored, length);
};

const char *cLinkWalker::LinksaveName(void)
{
  return *p_linksave_filename;
};



// --- Tools -----------------------------------------------------------------------

int ContainsSymlinks(const char *Directory) {
   struct stat st;
   if (stat(Directory, &st) != 0) {
      //esyslog("ERROR: Could not stat file: %s", Directory);
      LOG_ERROR_STR(Directory);
      return -1;
   } else if (!S_ISDIR(st.st_mode)) {
      esyslog("ERROR: Not a Directory: %s", Directory);
      return -1;;
   }
   
   int count = 0;
   cReadDir d(Directory);
   struct dirent *e;
   while ((e = d.Next()) != NULL) {
      char FileName[strlen(Directory) + strlen(e->d_name) + 5];
      sprintf(FileName, "%s/%s", Directory, e->d_name);
      dsyslog("Stat: %s", FileName);
      if (lstat(FileName, &st) == 0) {
         if (S_ISLNK(st.st_mode)) {
            count++;
         }
      } else {
         //esyslog("ERROR: Could not stat file: %s", FileName);
         LOG_ERROR_STR(FileName);
         return -1;
      }
   }
   return count;
}




cLastExistingDir::cLastExistingDir(const char *FileName)
{
   name = strdup(FileName);
   char *tmp = name;
   while ((tmp = strrchr(name,'/')) != NULL) {
      tmp[0] = 0;
      //dsyslog("processing directory: %s", name);
      struct stat st;
      if (stat(name, &st) == 0) {
         //dsyslog("stat directory: %s", name);
         if (S_ISDIR(st.st_mode)) {
            dsyslog("Last existing directory: %s", name);
            return;
         }  
      }
   }
   
   LOG_ERROR_STR(FileName);
   free(name);
   name = NULL;
   return;
};

cLastExistingDir::~cLastExistingDir()
{
   if (name)
      free(name);
};

const char *cLastExistingDir::Name(void)
{
  return name;
};

#endif
